import urllib.parse, typing

from .urlhandlers import UrlHandler, RequestsHandler
from .users import AdderUser
from .devices import AdderDevice, AdderReceiver, AdderTransmitter, AdderServer, AdderUSBExtender, AdderUSBReceiver, AdderUSBTransmitter
from .channels import AdderChannel
from .presets import AdderPreset

class AdderRequestError(Exception):
	"""Adder API request has not returned success"""
	pass

class AdderAPI:

	def __init__(self,server_address:str,*,url_handler:typing.Optional[UrlHandler]=None, user:typing.Optional[AdderUser]=None, api_version:typing.Optional[int]=8):
		"""Adderlink API for interacting with devices, channels, and users"""

		self._setServerAddress(server_address)
		self.setUrlHandler(url_handler or RequestsHandler())
		self.setUser(user or AdderUser())
		self.setApiVersion(api_version)
	
	def _setServerAddress(self, server_address:str):
		"""Set the server address to use"""

		# If no schema defined, indicate that properly in the address
		if "//" not in server_address:
			server_address = "//" + server_address
		
		self._server_address = urllib.parse.urlparse(server_address, scheme="http")

	def setUrlHandler(self, handler:UrlHandler):
		"""Set the UrlHandler to use for AdderAPI URL calls"""
		if not isinstance(handler, UrlHandler):
			raise ValueError(f"URL handler {type(handler)} is not an instance of UrlHandler")
		self._url_handler = handler

	def setUser(self, user:AdderUser):
		"""Set the AdderUser to use for AdderAPI calls"""
		if not isinstance(user, AdderUser):
			raise ValueError(f"User type{type(user)} is not an instance of AdderUser")
		self._user = user

	def setApiVersion(self, version:int):
		"""Set the API version to use"""
		self._api_version = int(version)

	# User authentication
	def login(self, username:str, password:str):
		"""Log the user in to the KVM system and retrieve an API token"""
		
		args = {
			"v":self._api_version,
			"method":"login",
			"username":username,
			"password":password
		}

		response = self._url_handler.api_call(self._server_address, args)
		
		if response.get("success") == "1" and response.get("token") is not None:
			self._user.set_logged_in(username, response.get("token"))
		
		elif "errors" in response:
			error = response.get("errors").get("error")
			raise AdderRequestError(f"Error {error.get('code','?')}: {error.get('msg','?')}")
		
		else:
			raise Exception("Unknown error")
	
	def logout(self):
		"""Log the user out"""

		args = {
			"v":self._api_version,
			"method":"logout",
			"token":self._user.token
		}

		response = self._url_handler.api_call(self._server_address, args)

		# TODO: More detailed error handling?
		# TODO: Maybe have the URL handler throw an exception?
		if response.get("success") == "1":
			self._user.set_logged_out()
		else:
			raise AdderRequestError()
		
	# Device management
	def getTransmitters(self, t_id:typing.Optional[str]=None) -> typing.Generator[AdderTransmitter, None, None]:
		"""Request a list of available Adderlink transmitters"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"get_devices",
			"device_type":"tx"
		}

		response = self._url_handler.api_call(self._server_address, args)

		if response.get("success") == "1" and "devices" in response:

			# It seems `xmltodict` only returns a list of nodes if there are more than one
			if response.get("count_devices") == "1":
				transmitters_list = [response.get("devices").get("device")]
			else:
				transmitters_list = response.get("devices").get("device")


			for device in transmitters_list:
				tx = AdderTransmitter(device)
				# Quick n dirty filtering since API does not support it natively
				if t_id is not None and t_id != tx.id:
					continue
				yield tx
			
	def getReceivers(self, r_id:typing.Optional[str]=None) -> typing.Generator[AdderReceiver, None, None]:
		"""Request a list of available Adderlink receivers"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"get_devices",
			"device_type":"rx"
		}

		response = self._url_handler.api_call(self._server_address, args)

		if response.get("success") == "1" and "devices" in response:

			# It seems `xmltodict` only returns a list of nodes if there are more than one
			if response.get("count_devices") == "1":
				receivers_list = [response.get("devices").get("device")]
			else:
				receivers_list = response.get("devices").get("device")

			for device in receivers_list:
				rx = AdderReceiver(device)
				# Quick n dirty filtering since API does not support it natively
				if r_id is not None and r_id != rx.id:
					continue
				yield rx

	# Device management
	def getServers(self) -> typing.Generator[AdderServer, None, None]:
		"""Request a list of available Adderlink AIM Servers"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"get_servers"
		}

		response = self._url_handler.api_call(self._server_address, args)

		if response.get("success") == "1" and "servers" in response:

			# It seems `xmltodict` only returns a list of nodes if there are more than one
			if response.get("count_servers") == "1":
				servers_list = [response.get("servers").get("server")]
			else:
				servers_list = response.get("servers").get("server")

			for server in servers_list:
				yield AdderServer(server)
	
	def setDeviceInfo(self, device:AdderDevice, *, description:typing.Optional[str]=None, location:typing.Optional[str]=None):
		"""Update the information for an existing device"""

		args = {
			"v": self._api_version,
			"token": self._user.token,
			"method": "update_device",
			"id": device.id
		}

		# Providing neither arguments will cause an API error anyway.  Might as well check here.
		if description is None and location is None:
			raise ValueError("At least one of `description` or `location` must be provided")

		# Set to underscore to properly clear an existing value (passing a blank string won't work)
		if description is not None:
			args.update({"desc": '_' if not len(description.strip()) else description})
		if location is not None:
			args.update({"loc": '_' if not len(location.strip()) else location})

		response = self._url_handler.api_call(self._server_address, args)			
			
		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			error = response.get("errors").get("error")
			raise AdderRequestError(f"Error {error.get('code','?')}: {error.get('msg','?')}")
		
		else:
			raise Exception("Unknown error")
	
	def rebootDevices(self, devices:typing.Union[typing.Iterable[AdderDevice], AdderDevice]):
		"""Reboot one or more devices"""
		
		devices = [devices] if isinstance(devices, AdderDevice) else devices

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"reboot_devices",
			"ids": ','.join(d.id for d in devices)
		}

		response = self._url_handler.api_call(self._server_address, args)

		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			error = response.get("errors").get("error")
			raise AdderRequestError(f"Error {error.get('code','?')}: {error.get('msg','?')}")

	
	# Channel management
	def getChannels(self, id:typing.Optional[str]=None, name:typing.Optional[str]="") -> typing.Generator[AdderChannel, None, None]:
		"""Request a list of available Adderlink channels"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"get_channels",
			"filter_c_name": name or ""
		}

		response = self._url_handler.api_call(self._server_address, args)
		
		if response.get("success") == "1" and "channels" in response:

			# It seems `xmltodict` only returns a list of nodes if there are more than one
			if response.get("count_channels") == "1":
				channel_list = [response.get("channels").get("channel")]
			else:
				channel_list = response.get("channels").get("channel")

			for channel in channel_list:
				ch = AdderChannel(channel)
				if id is not None and id != ch.id:
					continue
				yield AdderChannel(channel)
	
	def createChannel(self,
		name:str, description:typing.Optional[str]=None, location:typing.Optional[str]=None, modes:typing.Optional[list[AdderChannel.ConnectionMode]]=None, 
		video1:typing.Optional[AdderTransmitter]=None, video1_head:typing.Optional[int]=None,
		video2:typing.Optional[AdderTransmitter]=None, video2_head:typing.Optional[int]=None,
		audio:typing.Optional[AdderTransmitter]=None, usb:typing.Optional[AdderTransmitter]=None, serial:typing.Optional[AdderTransmitter]=None,
		group_name:typing.Optional[str]=None) -> AdderChannel:
		"""Create a new channel from the specified transmitters"""

		modes = modes or []
		
		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"create_channel",
			"name":name,
			"desc":description,
			"loc":location,
			"allowed":str().join([m.value for m in modes]).lower(),
			"groupname":group_name
		}

		# Video 1
		if video1:
			args.update({
				"video1":video1.id,
				"video1head":video1_head
			})
		
		# Video 2
		if video2:
			args.update({
				"video2":video2.id,
				"video2head":video2_head
			})
		
		# Audio
		if audio:
			args.update({"audio":audio.id})
		
		# USB
		if usb:
			args.update({"usb":usb.id})
		
		# Serial
		if serial:
			args.update({"serial":serial.id})
		
		response = self._url_handler.api_call(self._server_address, args)
		if response.get("success") == "1" and response.get("id"):
			return next(self.getChannels(id=response.get("id")))
		
		elif "errors" in response:
			error = response.get("errors").get("error")
			raise AdderRequestError(f"Error {error.get('code','?')}: {error.get('msg','?')}")
		
		else:
			raise Exception("Unknown error")


	def connectToChannel(self, channel:AdderChannel, receiver:AdderReceiver, mode:typing.Optional[AdderChannel.ConnectionMode]=AdderChannel.ConnectionMode.SHARED):
		"""Connect a channel to a receiver"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"connect_channel",
			"c_id":channel.id,
			"rx_id":receiver.id,
			"mode":mode.value
		}

		response = self._url_handler.api_call(self._server_address, args)
		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			error = response.get("errors").get("error")
			raise AdderRequestError(f"Error {error.get('code','?')}: {error.get('msg','?')}")
		
		else:
			raise Exception("Unknown error")
		
	def disconnectFromChannel(self, receiver:typing.Union[AdderReceiver, typing.Iterable[AdderReceiver]], force:typing.Optional[bool]=False):
		"""Disconnect a receiver -- or iterable of receivers -- from its current channel"""
		receiver = [receiver] if isinstance(receiver, AdderReceiver) else receiver

		args = {
			"v":self._api_version,
			"token": self._user.token,
			"method":"disconnect_channel",
			"rx_id":','.join(x.id for x in receiver),
			"force":int(force)
		}

		response = self._url_handler.api_call(self._server_address, args)
		
		if response.get("success") == "1":
			return

		elif "errors" in response:
			error = response.get("errors").get("error")
			raise AdderRequestError(f"Error {error.get('code','?')}: {error.get('msg','?')}")

	def deleteChannel(self, channel:AdderChannel):
		"""Delete a channel.  Admin privileges are required."""
		args = {
			"v": self._api_version,
			"token": self._user.token,
			"method": "delete_channel",
			"id": channel.id
		}
		
		response = self._url_handler.api_call(self._server_address, args)

		if response.get("success") == "1":
			return

		elif "errors" in response:
			error = response.get("errors").get("error")
			raise AdderRequestError(f"Error {error.get('code','?')}: {error.get('msg','?')}")		
	
	# Preset management
	def getPresets(self, id:typing.Optional[str]=None) -> typing.Generator[AdderPreset, None, None]:
		"""Request a list of available Adderlink presets"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"get_presets"
		}

		response = self._url_handler.api_call(self._server_address, args)
		
		if response.get("success") == "1" and "connection_presets" in response:

			# It seems `xmltodict` only returns a list of nodes if there are more than one
			if response.get("count_presets") == "1":
				presets_list = [response.get("connection_presets").get("connection_preset")]
			else:
				presets_list = response.get("connection_presets").get("connection_preset")

			for preset in presets_list:
				ps = AdderPreset(preset)
				if id is not None and id != ps.id:
					continue
				yield ps
	
	def createPreset(self, name:str, pairs:typing.Union[typing.Iterable[AdderPreset.Pair], AdderPreset.Pair], modes:typing.Union[typing.Iterable[AdderChannel.ConnectionMode], AdderChannel.ConnectionMode]) -> AdderPreset:
		"""Create a preset consisting of one or more channel/receiver pairs"""

		if not len(name.strip()):
			raise ValueError("'name' parameter must not be empty")
		name_formatted = urllib.parse.quote(name)

		if isinstance(pairs, AdderPreset.Pair):
			pairs = [pairs]
		pairs_formatted = ','.join(f"{pair.channel.id}-{pair.receiver.id}" for pair in pairs) # TODO: Handle 'ch.id-rx.id' formatting in Preset __str__?

		if isinstance(modes, AdderChannel.ConnectionMode):
			modes = [modes]
		modes_formatted = str().join(mode.value for mode in modes)

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"create_preset",
			"name":name_formatted,
			"pairs":pairs_formatted,
			"allowed":modes_formatted
		}

		response = self._url_handler.api_call(self._server_address, args)

		if response.get("success") == "1" and response.get("id"):
			return next(self.getPresets(response.get("id")))
		
		elif "errors" in response:
			raise Exception(f"Errors: {response.get('errors')}")
			#for error in response.get("errors").get("error"):
			#	raise Exception(f"Error {error.get('code','?')}: {error.get('msg','?')}")
	
	def loadPreset(self, preset:AdderPreset, mode:AdderChannel.ConnectionMode, force:typing.Optional[bool]=False):
		"""Connect a preset"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"connect_preset",
			"id":preset.id,
			"mode":mode.value,
			"force":int(force)
		}

		response = self._url_handler.api_call(self._server_address, args)
		
		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			raise Exception(f"Errors: {response.get('errors')}")
	
	def unloadPreset(self, preset:AdderPreset, force:typing.Optional[bool]=False):
		"""Disconnect a preset"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"disconnect_preset",
			"id":preset.id,
			"force":int(force)
		}

		response = self._url_handler.api_call(self._server_address, args)
		
		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			raise Exception(f"Errors: {response.get('errors')}")

	def deletePreset(self, preset:AdderPreset):
		"""Delete a preset"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"delete_preset",
			"id":preset.id
		}

		response = self._url_handler.api_call(self._server_address, args)
		
		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			raise Exception(f"Errors: {response.get('errors')}")
	
	# C-USB Lan Extender Management
	def getUSBReceivers(self, mac_address:typing.Optional[AdderUSBReceiver]=None) -> typing.Generator[AdderUSBReceiver, None, None]:
		"""Get a list of C-USB LAN Receivers, optionally filtered by MAC address"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"get_all_c_usb"
		}

		response = self._url_handler.api_call(self._server_address, args)

		if response.get("success") == "1" and "c_usb_lan_extenders" in response:

			# It seems `xmltodict` only returns a list of nodes if there are more than one
			if response.get("count_c_usbs") == "1":
				usb_list = [response.get("c_usb_lan_extenders").get("c_usb")]
			else:
				usb_list = response.get("c_usb_lan_extenders").get("c_usb")

			for usb in usb_list:
				if usb.get("type","") != "rx": continue

				rx = AdderUSBReceiver(usb)

				# Quick n dirty filtering since API does not support it natively
				if mac_address is not None and mac_address != rx.mac_address:
					continue

				yield rx

	# C-USB Lan Extender Management
	# TODO: Currently untested
	
	def getUSBTransmitters(self, mac_address:typing.Optional[AdderUSBTransmitter]=None) -> typing.Generator[AdderUSBTransmitter, None, None]:
		"""Get a list of C-USB LAN Transmitters, optionally filtered by MAC address"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"get_all_c_usb"
		}

		response = self._url_handler.api_call(self._server_address, args)

		if response.get("success") == "1" and "c_usb_lan_extenders" in response:

			# It seems `xmltodict` only returns a list of nodes if there are more than one
			if response.get("count_c_usbs") == "1":
				usb_list = [response.get("c_usb_lan_extenders").get("c_usb")]
			else:
				usb_list = response.get("c_usb_lan_extenders").get("c_usb")

			for usb in usb_list:
				if usb.get("type","") != "tx": continue

				tx = AdderUSBTransmitter(usb)

				# Quick n dirty filtering since API does not support it natively
				if mac_address is not None and mac_address != tx.mac_address:
					continue

				yield tx
	
	def deleteUSBExtender(self, usb:AdderUSBExtender):
		"""Delete a C-USB LAN Transmitter or Receiver"""

		args = {
			"v":self._api_version,
			"token":self._user.token,
			"method":"delete_c_usb",
			"mac":usb.mac_address
		}

		response = self._url_handler.api_call(self._server_address, args)
		
		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			raise Exception(f"Errors: {response.get('errors')}")
	
	def setUSBExtenderInfo(self, usb:AdderUSBExtender, *, name:str):
		"""Update the information for an existing USB extender"""

		args = {
			"v":self._api_version,
			"token":self._user._token,
			"method":"update_c_usb",
			"mac":usb.mac_address,
			"name":name
		}

		response = self._url_handler.api_call(self._server_address, args)
	
		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			raise Exception(f"Errors: {response.get('errors')}")


	def connectUSBExtender(self, receiver:AdderUSBReceiver, transmitter:AdderUSBTransmitter):
		"""Connect a C-USB LAN Receiver to a Transmitter"""

		args = {
			"v":self._api_version,
			"token":self._user._token,
			"method":"connect_c_usb",
			"rx": receiver.mac_address,
			"tx": transmitter.mac_address
		}

		response = self._url_handler.api_call(self._server_address, args)
	
		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			raise Exception(f"Errors: {response.get('errors')}")

	def disconnectUSBExtender(self, receiver:AdderUSBReceiver):
		"""Connect a C-USB LAN Receiver to a Transmitter"""

		args = {
			"v":self._api_version,
			"token":self._user._token,
			"method":"disconnect_c_usb",
			"mac": receiver.mac_address,
		}

		response = self._url_handler.api_call(self._server_address, args)
	
		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			raise Exception(f"Errors: {response.get('errors')}")
	
	def replaceDevice(self, old_device:AdderDevice, new_device:AdderDevice):
		"""Replace an old device with a new one"""

		args = {
			"v": self._api_version,
			"token": self._user._token,
			"method": "replace_device",
			"d_id": old_device.id,
			"r_d_id": new_device.id
		}

		response = self._url_handler.api_call(self._server_address, args)

		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			raise Exception(f"Errors: {response.get('errors')}")
	
	def identifyDevice(self, device:AdderDevice):
		"""Sends an 'identify' command to the specified device"""

		args = {
			"v": self._api_version,
			"token": self._user._token,
			"method": "identify_device",
			"id": device.id
		}

		response = self._url_handler.api_call(self._server_address, args)

		if response.get("success") == "1":
			return
		
		elif "errors" in response:
			raise Exception(f"Errors: {response.get('errors')}")

	@property
	def user(self) -> AdderUser:
		"""Get the current user"""
		return self._user

	@property
	def server_address(self) -> urllib.parse.ParseResult:
		"""Get the server address"""
		return self._server_address
	
	@property
	def url_handler(self) -> UrlHandler:
		"""Get the URL Handler"""
		return self._url_handler


import enum, abc, ipaddress, typing
from datetime import datetime
from dataclasses import dataclass

@dataclass
class NetworkInterface:
	"""Network interface info"""
	ip_address:typing.Union[ipaddress.IPv4Address,ipaddress.IPv6Address]
	mac_address:str
	is_online:bool


class AdderDevice(abc.ABC):
	"""Abstract Adder device"""
	@enum.unique
	class DeviceType(enum.Enum):
		"""Types of Adder devices"""
		RX = 0
		TX = 1
	
	@enum.unique
	class DeviceStatus(enum.Enum):
		"""Operational status of Adder device"""
		UNKNOWN         = -1
		OFFLINE         =  0
		ONLINE          =  1
		REBOOTING       =  2
		UPGRADING       =  4
		BACKUP_FIRMWARE =  6
	
	@enum.unique
	class DeviceModel(enum.Enum):
		"""AdderLink device models"""
		UNKNOWN  =  '?'
		ALIF1000 =  '1'
		ALIF2002 =  'b'
		ALIF2112 =  'v'
		ALIF1002 =  's'
		ALIF2020 =  't'
		ALIF100T =  'd'
		ALIF100T_VGA = 'f',
		ALIF101T_HDMI = 'h',
		ALIF4021R = '4',
		ALIFE300R = '8'
	
	
	def __init__(self, properties:dict):
		# Do a lil copy
		self._extended = {key: val for key,val in properties.items()}
	
	# Human-friendly descriptions
	@property
	def name(self) -> str:
		"""Device name"""
		return self._extended.get("d_name","")

	@property
	def description(self) -> str:
		"""Device description"""
		return self._extended.get("d_description","")

	@property
	def location(self) -> str:
		"""Device location"""
		return self._extended.get("d_location","")
	
	@property
	def date_dadded(self) -> datetime:
		"""The date/time this device was set up"""
		return datetime.fromisoformat(self._extended.get("d_date_added"))
	
	# Addresses and IDs
	@property
	def id(self) -> str:
		"""Device ID"""
		return self._extended.get("d_id","")
	
	@property
	def serial_number(self) -> str:
		"""Device serial number"""
		return self._extended.get("d_serial_number","")
	
	@property
	def ip_addresses(self) -> typing.Tuple[typing.Union[ipaddress.IPv4Address,ipaddress.IPv6Address,None]]:
		"""IP addresses on the network"""
		return (
			ipaddress.ip_address(self._extended.get("d_ip_address")) if self._extended.get("d_ip_address") else None,
			ipaddress.ip_address(self._extended.get("d_ip_address2")) if self._extended.get("d_ip_address2") else None,
		)
	
	@property
	def ip_address(self) -> typing.Union[ipaddress.IPv4Address,ipaddress.IPv6Address,None]:
		"""Primary IP address of the device"""
		return self.ip_addresses[0] or None

	@property
	def mac_addresses(self) -> typing.Tuple[str]:
		"""MAC addresses of the interfaces"""
		return (self._extended.get("d_mac_address",""), self._extended.get("d_mac_address2",""))
	
	@property
	def mac_address(self) -> str:
		"""Primary MAC address"""
		return self.mac_addresses[0]

	@property
	def network_interfaces(self) -> typing.Tuple[NetworkInterface]:
		"""Return network interfaces"""
		return (
			NetworkInterface(self.ip_addresses[0], self.mac_addresses[0], int(self._extended.get("d_online",-1))>0),
			NetworkInterface(self.ip_addresses[1], self.mac_addresses[1], int(self._extended.get("d_online2",-1))>0)
		)

	# Firmware verions
	@property
	def firmware(self) -> str:
		"""Current firmware version"""
		return self._extended.get("d_firmware","")
	
	@property
	def backup_firmware(self) -> str:
		"""Backup firmware version"""
		return self._extended.get("d_backup_firmware","")
	
	# Status
	@property
	def status(self) -> DeviceStatus:
		"""Device status"""
		try:
			return self.DeviceStatus(int(self._extended.get("d_status")))
		except Exception as e:
			return self.DeviceStatus(-1)

	@property
	def is_configured(self) -> bool:
		"""If the device has been set up"""
		return int(self._extended.get("d_configured",0)) > 0
	
	@property
	def is_firmware_valid(self) -> bool:
		"""If the current firmware is in working order"""
		return int(self._extended.get("d_valid_firmware")) > 0
	
	@property
	def is_backup_firmware_valid(self) -> bool:
		"""If the current firmware is in working order"""
		return int(self._extended.get("d_valid_backup_firmware")) > 0
	
	@property
	def model(self) -> DeviceModel:
		"""Device model"""
		try:
			if self._extended.get("d_version") == '2':
				return self.DeviceModel(self._extended.get("d_variant"))
			else:
				return self.DeviceModel(self._extended.get("d_version"))
		except:
			return self.DeviceModel.UNKNOWN

	def __repr__(self):
		return f"<{self.__class__.__name__} name=\"{self.name}\" location=\"{self.location}\" model={self.model.name} id={self.id}>"


class AdderTransmitter(AdderDevice):
	"""Adderlink Transmitter (TX) Device"""

	@property
	def channel_count(self) -> int:
		"""Number of channels that use this device"""
		return int(self._extended.get("count_transmitter_channels",0))
	
	@property
	def preset_count(self) -> int:
		"""Number of presets that use this device"""
		return int(self._extended.get("count_transmitter_presets",0))


class AdderReceiver(AdderDevice):
	"""Adderlink Receiver (RX) Device"""

	# TODO: con_exclusive(bool) vs count_control(enum) for determining exclusive mode?

	@enum.unique
	class ConnectionControlType(enum.Enum):
		"""Control modes"""
		UNKNOWN    = -1
		VIDEO_ONLY =  1
		EXCLUSIVE  =  2
		SHARED     =  3
	
	# Connection info

	@property
	def connection_start(self) -> typing.Optional[datetime]:
		"""Time the last known connection started"""
		if self._extended.get("con_start_time"):
			return datetime.fromisoformat(self._extended.get("con_start_time"))
		else:
			return None
		
	@property
	def connection_end(self) -> typing.Optional[datetime]:
		"""Time the last known connection was ended.  Returns None if connection is current."""
		if self._extended.get("con_end_time"):
			return datetime.fromisoformat(self._extended.get("con_end_time"))
		else:
			return None

	@property
	def connection_control(self) -> ConnectionControlType:
		"""Control mode of the last known connection"""
		con_type = int(self._extended.get("con_control", -1))
		try:
			return self.ConnectionControlType(con_type)
		except Exception:
			return self.ConnectionControlType(-1)
	
	@property
	def channel_name(self) -> str:
		"""The name of the last known channel this receiver was connected"""
		return self._extended.get("c_name","")

	@property
	def is_connected(self) -> bool:
		"""Is the receiver currently connected to a channel"""
		return self.connection_start and not self.connection_end
	
	# Connected user info
	@property
	def last_username(self) -> str:
		"""Last known username"""
		return self._extended.get("u_username","")
	
	@property
	def last_userid(self) -> str:
		"""Last known user ID"""
		print(self._extended)
		return self._extended.get("u_id","")
	
	@property
	def current_username(self) -> str:
		"""Name of current user connected"""
		return self.last_username if self.is_connected else ""

	@property
	def current_userid(self) -> str:
		"""User ID of current user connected"""
		return self.last_userid if self.is_connected else ""
	
	# Stats
	@property
	def group_count(self) -> int:
		"""Number of receiver groups this belongs to"""
		return int(self._extended.get("count_receiver_groups",0))

	@property
	def preset_count(self) -> int:
		"""Number of receiver presets this belongs to"""
		return int(self._extended.get("count_receiver_presets",0))
	
	@property
	def user_count(self) -> int:
		"""Number of users with access to this receiver"""
		return int(self._extended.get("count_users",0)) 

class AdderUSBExtender(abc.ABC):
	"""Abstract Adder C-USB LAN Extender Device"""
	
	def __init__(self, properties:dict):
		# Do a lil copy
		self._extended = {key: val for key,val in properties.items()}
	
	@property
	def name(self) -> str:
		"""Device name"""
		return self._extended.get("d_name","")
	
	@property
	def ip_address(self) -> typing.Union[ipaddress.IPv4Address, ipaddress.IPv6Address, None]:
		return ipaddress.ip_address(self._extended.get("ip")) if self._extended.get("ip") else None

	@property
	def mac_address(self) -> str:
		return self._extended.get("mac","")
	
	@property
	def is_online(self) -> bool:
		return self._extended.get("online") == 1
	
	@property
	def network_interface(self) -> NetworkInterface:
		return NetworkInterface(self.ip_address, self.mac_address, self.is_online)

class AdderUSBTransmitter(AdderUSBExtender):
	"""Adder C-USB LAN Network Transmitter"""

class AdderUSBReceiver(AdderUSBExtender):
	"""Adder C-USB Lan Network Receiver"""

	@property
	def connected_to(self) -> str:
		"""MAC address of the connected transmitter"""
		return self._extended.get("connectedTo","")

class AdderServer:
	"""An Adder Server Device"""

	@enum.unique
	class Role(enum.Enum):
		"""The role of the AIM"""
		SOLO    = "solo",
		BACKUP  = "backup",
		PRIMARY = "primary",
		UNCONFIGURED = "unconfigured",
		UNKNOWN = "unknown"
	
	@enum.unique
	class Status(enum.Enum):
		"""The status of the AIM"""
		ACTIVE   = "active",
		STANDBY  = "standby",
		FAILED   = "failed",
		QUISCENT =  "quiscent",
		UNKNOWN  = "unknown"
	
	@enum.unique
	class DualEthernetConfig(enum.Enum):
		"""The configuration of the second ethernet port"""
		NO      = 0,
		DHCP    = 1,
		STATIC  = 2,
		BONDED  = 3,
		UNKNOWN = -1


	def __init__(self, properties:dict):
		# Do a lil copy
		self._extended = {key: val for key,val in properties.items()}
	
	# Human-friendly descriptions
	@property
	def name(self) -> str:
		"""Server name"""
		return self._extended.get("name","")

	@property
	def description(self) -> str:
		"""Server description"""
		return self._extended.get("description","")

	@property
	def location(self) -> str:
		"""Server location"""
		return self._extended.get("location","")
	
	@property
	def role(self) -> Role:
		"""The configured role of the AIM"""
		try:
			return self.Role(self._extended.get("role"))
		except Exception:
			return self.Role("unknown")
	
	@property
	def status(self) -> Status:
		"""The current status of the AIM"""
		try:
			return self.Status(self._extended.get("status"))
		except Exception:
			return self.Status("unknown")
	
	@property
	def ip_addresses(self) -> typing.Tuple[typing.Union[ipaddress.IPv4Address,ipaddress.IPv6Address,None]]:
		"""IP addresses on the network"""
		return (
			ipaddress.ip_address(self._extended.get("ip")) if self._extended.get("ip") else None,
			ipaddress.ip_address(self._extended.get("ip2")) if self._extended.get("ip2") else None,
		)
	
	@property
	def ip_address(self) -> typing.Union[ipaddress.IPv4Address,ipaddress.IPv6Address,None]:
		"""Primary IP address of the device"""
		return self.ip_addresses[0] or None

	@property
	def mac_addresses(self) -> typing.Tuple[str]:
		"""MAC addresses of the interfaces"""
		return (self._extended.get("mac",""), self._extended.get("mac2",""))
	
	@property
	def mac_address(self) -> str:
		"""Primary MAC address"""
		return self.mac_addresses[0]

	@property
	def dual_ethernet_config(self) -> DualEthernetConfig:
		"""The configuration of the second ethernet port"""
		try:
			return self.DualEthernetConfig(int(self._extended.get("eth1",-1)))
		except Exception:
			return self.DualEthernetConfig(-1)


import enum


class AdderChannel:
	"""Adderlink channel"""

	@enum.unique
	class ConnectionMode(enum.Enum):
		VIEW_ONLY = 'v'
		SHARED    = 's'
		EXCLUSIVE = 'e'
		PRIVATE   = 'p'

	@enum.unique
	class ButtonState(enum.Enum):
		"""Video-only mode capabilities"""
		UNKNOWN  = "unknown"  # Unknown state
		DISABLED = "disabled" # Disabled because in use
		ENABLED  = "enabled"  # Enabled
		HIDDEN   = "hidden"   # Never allowed

	def __init__(self, properties:dict):
		self._extended = {key: val for key,val in properties.items()}
	
	@property
	def id(self) -> str:
		"""Channel ID"""
		return self._extended.get("c_id","")
	
	@property
	def name(self) -> str:
		"""Channel name"""
		return self._extended.get("c_name","")
	
	@property
	def description(self) -> str:
		"""Channel description"""
		return self._extended.get("c_description","")
	
	@property
	def location(self) -> str:
		"""Channel location"""
		return self._extended.get("c_location","")
	
	@property
	def type(self) -> str:
		"""Channel type"""
		# TODO: Investigate.  Not in the documentation
		return self._extended.get("c_channel_type","")

	@property
	def tx_id(self) -> str:
		"""Device ID"""
		# TODO: Investigate.  Not in the documentation
		return self._extended.get("c_tx_id","")
	
	@property
	def is_online(self) -> bool:
		"""Device status"""
		# TODO: Investigate.  Not in the documentation
		return self._extended.get("channel_online") == "1"
	
	@property
	def is_favorite(self) -> bool:
		"""Whether the channel has been favorited"""
		return self._extended.get("c_favourite") != "false"
	
	@property
	def shortcut(self) -> str:
		"""Returns the shortcut index if set"""
		return self._extended.get("c_favourite","") if self.is_favorite else ''
	
	@property
	def view_button(self) -> ButtonState:
		"""Indicates the state of the video-only view button"""
		state = self._extended.get("view_button")
		try:
			return self.ButtonState(state)
		except ValueError:
			return self.ButtonState.UNKNOWN
	
	@property
	def shared_button(self) -> ButtonState:
		"""Indicates the state of the shared view button"""
		state = self._extended.get("shared_button")
		try:
			return self.ButtonState(state)
		except ValueError:
			return self.ButtonState.UNKNOWN

	@property
	def control_button(self) -> ButtonState:
		"""Indicates the state of the full-control button"""
		state = self._extended.get("control_button")
		try:
			return self.ButtonState(state)
		except ValueError:
			return self.ButtonState.UNKNOWN

	@property
	def exclusive_button(self) -> ButtonState:
		"""Indicates the state of the video-only view button"""
		state = self._extended.get("exclusive_button")
		try:
			return self.ButtonState(state)
		except ValueError:
			return self.ButtonState.UNKNOWN
	
	@property
	def view_available(self) -> bool:
		"""Whether the user may view in video-only mode"""
		return self.view_button == self.ButtonState.ENABLED

	@property
	def shared_available(self) -> bool:
		"""Whether the user may view in shared mode"""
		return self.shared_button == self.ButtonState.ENABLED
	
	@property
	def control_available(self) -> bool:
		"""Whether the user may view in video-only mode"""
		return self.control_button == self.ButtonState.ENABLED
	
	@property
	def exclusive_available(self) -> bool:
		"""Whether the user may view in video-only mode"""
		return self.exclusive_button == self.ButtonState.ENABLED
	
	
	def _print_undocumented(self):
		"""Undocumented features"""
		for property in ["c_video1","c_video1_head","c_video2","c_video2_head","c_audio","c_usb","c_serial","c_usb1", "c_audio1","c_audio2","c_sensitive","c_rdp_id"]:
			print(f"{property.ljust(16)}: {self._extended.get(property,'Not Set')}")
		


	"""
	TODO: Additional values not documented

	Additional channel output values in version 4:
	- c_video1 (device ID)
	- c_video1_head (1|2)
	- c_video2 (device ID)
	- c_video2_head (1|2)
	- c_audio (device ID)
	- c_usb (device ID)
	- c_serial (device ID)

	Additional channel output values in version 8:
	- c_usb1 (device ID)
	- c_audio1 (device ID)
	- c_audio2 (device ID)
	- c_sensitive
	- c_rdp_id (RDP ID) only for RDP devices.
	"""

	def __repr__(self):
		return f"<{self.__class__.__name__} name=\"{self.name}\" is_online={self.is_online} id={self.id}>"


import abc, urllib.parse
import requests
import xmltodict

#class InvalidServerAddressError(RuntimeError):
#	"""The server address provided is missing or invalid"""

class UrlHandler(abc.ABC):
	"""Abstract URL Handler"""

	@abc.abstractclassmethod
	def api_call(cls, server_address:urllib.parse.ParseResult, args:dict) -> dict:
		"""
		Handle a call to the REST API and return a dictionary result
		Data from the Adder API is returned as an XML document and should be returned as a dictionary
		_parse_response() is provided as a good-enough method via xmltodict, but can be overloaded if desired
		"""
		pass

	@classmethod
	def _parse_response(cls, data:str) -> xmltodict.OrderedDict:
		"""Parse an API response to a Python data structure"""
		return xmltodict.parse(data).get("api_response")

	@classmethod
	def _build_url(cls, server_address:urllib.parse.ParseResult, args:dict) -> str:
		"""Build a full URL for an API request"""
		return urllib.parse.urljoin(server_address.geturl(), f"/api/?{urllib.parse.urlencode(args)}")

class RequestsHandler(UrlHandler):
	"""Request handler using the `requests` library"""

	timeout:int=5

	@classmethod
	def api_call(cls, server_address:urllib.parse.ParseResult, args:dict) -> xmltodict.OrderedDict:
		"""GET a call to the API"""

		response = requests.get(cls._build_url(server_address, args), timeout=cls.timeout)
		
		if not response.ok:
			raise Exception(f"Error contacting {server_address.netloc}: Returned {response.status_code}")

		return cls._parse_response(response.content)


class DebugHandler(UrlHandler):
	"""URL handler for debugging with sample XMLs from the documentation"""
	
	@classmethod
	def api_call(cls, server_address:urllib.parse.ParseResult, args:dict) -> xmltodict.OrderedDict:
		"""Load sample XML return data for the given query"""

		full_url = cls._build_url(server_address, args)
		print(f"Requesting: {full_url}")
		print(f"With params: {args}")
		
		method = args.get("method")[0]

		if method == "login":
			if args.get("password")[0] == "goodpwd":
				with open("example_xml/login.xml") as api_response:
					response = cls._parse_response(api_response.read())
				return response
			else:
				with open("example_xml/login_fail.xml") as api_response:
					response = cls._parse_response(api_response.read())
				return response
		
		if method == "logout":
			with open("example_xml/logout.xml") as api_response:
				response = cls._parse_response(api_response.read())
			return response
		
		elif method == "get_devices":
			with open("example_xml/get_devices.xml") as api_response:
				response = cls._parse_response(api_response.read())
			#	print("Responding with",response)
			return response
		
		elif method == "get_channels":
			with open("example_xml/get_channels.xml") as api_response:
				response = cls._parse_response(api_response.read())
			#	print("Responding with",response)
			return response


class AdderUser:
	"""Adder user account"""

	def __init__(self):
		self._username = ""
		self._token = ""
	
	def set_logged_in(self, user:str, token:str):
		"""Log a user in with a valid session token"""
		self._username = user
		self._token = token
	
	def set_logged_out(self):
		"""Log the user out"""
		self._username = ""
		self._token = ""

	@property
	def logged_in(self) -> bool:
		"""Whether the user has been successfully authenticated"""
		return self._username and self._token
	
	@property
	def logged_out(self) -> bool:
		"""Whether the use is not logged in"""
		return not self.logged_in
	
	@property
	def username(self) -> str:
		"""The username for this user"""
		return self._username
	
	@property
	def token(self) -> str:
		"""The authentication token for the current session"""
		return self._token


import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# Thanks https://realpython.com/pypi-publish-python-package/
setup(
    name="adderlib",
    version="1.0.1",
    description="Python wrapper for the Adder API, for use with Adderlink KVM systems",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://adderlib.readthedocs.io/",
    author="Michael Jordan",
    author_email="michael@glowingpixel.com",
    license="GNU General Public License v3 or later (GPLv3+)",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
		"Development Status :: 4 - Beta",
		"Intended Audience :: Developers",
		"Topic :: Multimedia",
		"Topic :: Internet",
		"Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=["adderlib"],
    include_package_data=True,
    install_requires=["requests", "xmltodict"]
)


import enum, dataclasses
from .devices import AdderReceiver
from .channels import AdderChannel

class AdderPreset:
	"""Adderlink Preset"""

	@dataclasses.dataclass
	class Pair:
		"""A channel/receiver pair"""
		channel:  AdderChannel
		receiver: AdderReceiver

	@enum.unique
	class ActiveState(enum.Enum):
		FULL    = "full"
		PARTIAL = "partial"
		NONE    = "none"
		UNKNOWN = "unknown"

	@enum.unique
	class ButtonState(enum.Enum):
		"""Video-only mode capabilities"""
		UNKNOWN  = "unknown"  # Unknown state
		DISABLED = "disabled" # Disabled because in use
		ENABLED  = "enabled"  # Enabled
		HIDDEN   = "hidden"   # Never allowed

	def __init__(self, properties:dict):
		self._extended = {key: val for key,val in properties.items()}
	
	@property
	def id(self) -> str:
		"""Preset ID"""
		return self._extended.get("cp_id","")
	
	@property
	def name(self) -> str:
		"""Preset name"""
		return self._extended.get("cp_name","")
	
	@property
	def description(self) -> str:
		"""Preset description"""
		return self._extended.get("cp_description","")
	
	@property
	def pair_count(self) -> int:
		"""Number of valid channel/receiver pairs"""
		cp_pairs = self._extended.get("cp_pairs","")
		return int(cp_pairs) if cp_pairs.isnumeric() else 0
	
	@property
	def pair_problem_count(self) -> int:
		"""Number of problematic channel/receiver pairs"""
		problem_cp_pairs = self._extended.get("problem_cp_pairs","0")
		return int(problem_cp_pairs) if problem_cp_pairs.isnumeric() else 0
	
	@property
	def currently_active(self) -> ActiveState:
		"""Determines if any of the channel/receiver pairs are currently active"""
		try:
			return self.ActiveState(self._extended.get("cp_active",""))
		except ValueError:
			return self.ActiveState.UNKNOWN

	@property
	def connected_rx_count(self) -> int:
		"""The number of receivers already connected to this preset"""
		rx_count = self._extended.get("connected_rx_count","")
		return int(rx_count) if rx_count.isnumeric() else 0

	@property
	def view_button(self) -> ButtonState:
		"""Indicates the state of the video-only view button"""
		state = self._extended.get("view_button")
		try:
			return self.ButtonState(state)
		except ValueError:
			return self.ButtonState.UNKNOWN
	
	@property
	def shared_button(self) -> ButtonState:
		"""Indicates the state of the shared view button"""
		state = self._extended.get("shared_button")
		try:
			return self.ButtonState(state)
		except ValueError:
			return self.ButtonState.UNKNOWN

	@property
	def control_button(self) -> ButtonState:
		"""Indicates the state of the full-control button"""
		state = self._extended.get("control_button")
		try:
			return self.ButtonState(state)
		except ValueError:
			return self.ButtonState.UNKNOWN

	@property
	def exclusive_button(self) -> ButtonState:
		"""Indicates the state of the video-only view button"""
		state = self._extended.get("exclusive_button")
		try:
			return self.ButtonState(state)
		except ValueError:
			return self.ButtonState.UNKNOWN